from enum import Enum

from boto3.dynamodb.types import TypeDeserializer

from wellcome_aws_utils.exceptions import UnWellcomeException


def create_dynamo_events(event):
    for record in event['Records']:
        yield DynamoEvent(record)


class DynamoEventType(Enum):
    REMOVE, INSERT, MODIFY = range(3)


class DynamoEvent:
    def _set_event_type(self, record):
        if self.record['eventName'] == 'REMOVE':
            self.event_type = DynamoEventType.REMOVE
        elif self.record['eventName'] == 'INSERT':
            self.event_type = DynamoEventType.INSERT
        elif self.record['eventName'] == 'MODIFY':
            self.event_type = DynamoEventType.MODIFY
        else:
            raise UnWellcomeException(
                f'Unrecognised eventName found in {record}!'
            )

    def __init__(self, record):
        self.record = record

        try:
            self._set_event_type(record)
            self.event_source_arn = record['eventSourceARN']

            self._keys = record['dynamodb']['Keys']
            self._new_image = record['dynamodb'].get('NewImage')
            self._old_image = record['dynamodb'].get('OldImage')
        except KeyError as e:
            raise UnWellcomeException(
                f'{e} not found in {record}!'
            )

    @staticmethod
    def _deserialize_values(image):
        td = TypeDeserializer()
        return {k: td.deserialize(v) for k, v in image.items()}

    def keys(self, deserialize_values=False):
        if deserialize_values and self._keys:
            return DynamoEvent._deserialize_values(self._keys)

        return self._keys

    def new_image(self, deserialize_values=False):
        if deserialize_values and self._new_image:
            return DynamoEvent._deserialize_values(self._new_image)

        return self._new_image

    def old_image(self, deserialize_values=False):
        if deserialize_values and self._old_image:
            return DynamoEvent._deserialize_values(self._old_image)

        return self._old_image



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
get records from VHS, apply the transformation to them, and shove them into
an elasticsearch index
"""
import json
import boto3
import certifi
from attr import attrs, attrib
from elasticsearch import Elasticsearch
from wellcome_aws_utils.lambda_utils import log_on_error


def get_es_credentials(profile_name=None):
    session = boto3.session.Session(profile_name=profile_name)
    client = session.client(
        service_name='secretsmanager',
        region_name="eu-west-1"
    )
    get_secret_value_response = client.get_secret_value(
        SecretId="prod/Elasticsearch/ReportingCredentials"
    )
    secret = get_secret_value_response['SecretString']
    return json.loads(secret)


def dict_to_location(d):
    return ObjectLocation(**d)


@attrs
class ObjectLocation(object):
    namespace = attrib()
    path = attrib()


@attrs
class Record(object):
    id = attrib()
    version = attrib()
    payload = attrib(converter=dict_to_location)


@attrs
class ElasticsearchRecord(object):
    id = attrib()
    doc = attrib()


def extract_sns_messages_from_event(event):
    keys_to_keep = ['id', 'version', 'payload']

    for record in event["Records"]:
        try:
            full_message = json.loads(record["Sns"]["Message"])
        except KeyError:
            # This could be a message from SQS rather than SNS so attempt to
            # decode the "body" of the message
            event = json.loads(record["body"])
            for msg in extract_sns_messages_from_event(event):
                yield msg
        else:
            stripped_message = {
                k: v for k, v in full_message.items() if k in keys_to_keep
            }
            yield stripped_message


def get_dynamo_record(dynamo_table, message):
    item = dynamo_table.get_item(Key={"id": message['id']})
    return Record(**item["Item"])


def get_s3_objects_from_messages(dynamo_table, s3, messages):
    for message in messages:
        record = get_dynamo_record(dynamo_table, message)
        s3_object = s3.get_object(
            Bucket=record.payload.namespace,
            Key=record.payload.path
        )
        yield record.id, s3_object


def unpack_json_from_s3_objects(s3_objects):
    for id, s3_object in s3_objects:
        data = s3_object["Body"].read().decode("utf-8")
        yield id, json.loads(data)


def transform_data_for_es(data, transform):
    for id, data_dict in data:
        yield ElasticsearchRecord(
            id=id,
            doc=transform(data_dict)
        )


@log_on_error
def process_messages(
    event, transform, index, table_name, dynamodb=None, s3_client=None,
    es_client=None, credentials=None
):
    s3_client = s3_client or boto3.client("s3")
    dynamo_table = (dynamodb or boto3.resource("dynamodb")).Table(table_name)

    if credentials and not es_client:
        es_client = Elasticsearch(
            hosts=credentials["url"],
            use_ssl=True,
            ca_certs=certifi.where(),
            http_auth=(credentials['username'], credentials['password'])
        )

    elif not es_client:
        raise ValueError(
            'process_messages needs an elasticsearch client or a set of '
            'credentials to create one'
        )

    _process_messages(
        event, transform, index, dynamo_table, s3_client, es_client
    )


def _process_messages(
    event, transform, index, dynamo_table, s3_client, es_client
):
    messages = extract_sns_messages_from_event(event)
    s3_objects = get_s3_objects_from_messages(
        dynamo_table, s3_client, messages
    )
    data = unpack_json_from_s3_objects(s3_objects)
    es_records_to_send = transform_data_for_es(data, transform)

    for record in es_records_to_send:
        es_client.index(
            index=index,
            doc_type="_doc",
            id=record.id,
            body=json.dumps(record.doc)
        )



# -*- encoding: utf-8 -*-

import collections
import datetime
import decimal
import json
import logging
import warnings

from wellcome_aws_utils.exceptions import UnWellcomeException

SNSEvent = collections.namedtuple('SNSEvent', 'subject message')


logger = logging.getLogger(__name__)


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()

        if isinstance(obj, decimal.Decimal):
            if float(obj).is_integer():
                return int(obj)
            else:
                return float(obj)

        return json.JSONEncoder.default(self, obj)


def publish_sns_message(sns_client,
                        topic_arn,
                        message,
                        subject="default-subject"):
    """
    Given a topic ARN and a series of key-value pairs, publish the key-value
    data to the SNS topic.
    """
    response = sns_client.publish(
        TopicArn=topic_arn,
        MessageStructure='json',
        Message=json.dumps({
            'default': json.dumps(
                message,
                cls=EnhancedJSONEncoder
            )
        }),
        Subject=subject
    )

    if response['ResponseMetadata']['HTTPStatusCode'] == 200:
        logger.debug('SNS: sent notification %s', response["MessageId"])
    else:
        raise RuntimeError(repr(response))

    return response


def extract_sns_messages_from_lambda_event(event):
    """
    Extracts a JSON message from an SNS event sent to an AWS Lambda.

    :param event: An event sent to a Lambda from SNS.
    :returns: A generator of SNSEvent instances.

    """
    if 'Records' not in event:
        raise UnWellcomeException(f'No records found in {event}')

    for record in event['Records']:
        if record['EventSource'] != 'aws:sns':
            raise UnWellcomeException(f'Invalid message source for {record}')

        try:
            subject = record['Sns']['Subject']
            message = json.loads(record['Sns']['Message'])
        except KeyError as e:
            raise UnWellcomeException(
                f'Invalid message structure, missing {e} in {record}'
            )

        yield SNSEvent(subject=subject, message=message)


def extract_json_message(event):
    """
    Extracts a JSON message from an SNS event sent to a lambda

    Deprecated in favour of extract_sns_messages_from_lambda_event
    """
    warnings.warn(
        'Deprecated in favour of extract_sns_messages_from_lambda_event',
        DeprecationWarning
    )

    message = event['Records'][0]['Sns']['Message']
    return json.loads(message)



# -*- encoding: utf-8 -*-

import json
from urllib.parse import unquote

import boto3
from botocore.exceptions import ClientError
import dateutil.parser


def is_object(bucket, key):
    """
    Checks if an object exists in S3.  Returns True/False.

    :param bucket: Bucket of the object to check.
    :param key: Key of the object to check.

    """
    client = boto3.client('s3')
    try:
        client.head_object(Bucket=bucket, Key=key)
    except ClientError as err:
        if err.response['Error']['Code'] == '404':
            return False
        else:
            raise
    else:
        return True


def copy_object(src_bucket, src_key, dst_bucket, dst_key, lazy=False):
    """
    Copy an object from one S3 bucket to another.

    :param src_bucket: Bucket of the source object.
    :param src_key: Key of the source object.
    :param dst_bucket: Bucket of the destination object.
    :param dst_key: Key of the destination object.
    :param lazy: Do a lazy copy.  This means that the object will only be
        copied if the destination object does not exist, or exists but has
        a different ETag from the source object.

    """
    client = boto3.client('s3')
    if not is_object(bucket=src_bucket, key=src_key):
        raise ValueError(
            f'Tried to copy missing object ({src_bucket}, {src_key})'
        )

    def should_copy():
        if not lazy:
            return True

        if not is_object(bucket=dst_bucket, key=dst_key):
            return True

        src_resp = client.head_object(Bucket=src_bucket, Key=src_key)
        dst_resp = client.head_object(Bucket=dst_bucket, Key=dst_key)
        return src_resp['ETag'] == dst_resp['ETag']

    if should_copy():
        return client.copy_object(
            CopySource={'Bucket': src_bucket, 'Key': src_key},
            Bucket=dst_bucket,
            Key=dst_key
        )


def _extract_s3_event(record):
    event_datetime = dateutil.parser.parse(record["eventTime"])

    return {
        "event_name": record["eventName"],
        "event_time": event_datetime,
        "bucket_name": record["s3"]["bucket"]["name"],
        "object_key": unquote(record["s3"]["object"]["key"]),
        "size": record["s3"]["object"]["size"],
        "versionId": record["s3"]["object"].get("versionId")
    }


def parse_s3_record(event):
    """
    Extracts a simple subset of an S3 update event.
    """
    return [_extract_s3_event(record) for record in event["Records"]]


def write_objects_to_s3(bucket, key, objects):
    """
    Given an iterable of objects that can be serialised as JSON, serialise
    them as JSON, and write them to a file in S3, one per line.

    :param bucket: S3 bucket to upload the new file to.
    :param key: S3 key to upload the new file to.
    :param objects: An iterable of objects that can be serialised as JSON.

    """
    # We use sort_keys=True to ensure deterministic results.  The separators
    # flag allows us to write more compact JSON, which makes things faster!
    # See https://twitter.com/raymondh/status/842777864193769472
    json_str = b'\n'.join([
        json.dumps(m, sort_keys=True, separators=(',', ':')).encode('ascii')
        for m in objects
    ])

    client = boto3.client('s3')
    client.put_object(Bucket=bucket, Key=key, Body=json_str)


__all__ = [
    'is_object',
    'copy_object',
    'parse_s3_record',
    'write_objects_to_s3',
]



# -*- encoding: utf-8 -*-

import operator

from botocore.exceptions import ClientError


class EcsThrottleException(Exception):
    pass


def identify_cluster_by_app_name(client, app_name):
    """
    Given the name of one of our applications (e.g. api, calm_adapter),
    return the ARN of the cluster the task runs on.
    """
    for cluster_arn in get_cluster_arns(client):
        for service_arn in get_service_arns(client, cluster_arn=cluster_arn):
            # The format of an ECS service ARN is:
            #
            #     arn:aws:ecs:{aws_region}:{account_id}:service/{service_name}
            #
            # Our ECS cluster is configured so that the name of the ECS cluster
            # matches the name of the config in S3.  It would be more robust
            # to use the describeService API, but this saves us a couple of
            # calls on our API quota so we skip it.
            _, service_name = service_arn.split('/')
            if service_name == app_name:
                return cluster_arn

    raise RuntimeError(f'Unable to find ECS cluster for {app_name}')


def get_latest_task_definition(client, cluster, service):
    """
    Given the name of a cluster and a service, return the ARN
    for its latest task definition.
    """
    resp = client.describe_services(cluster=cluster, services=[service])

    # The top-level structure of a describeServices API response is of the form
    #
    #     {
    #       "failures": [],
    #       "services": [
    #         ...
    #       ]
    #     }
    #
    # Because we only asked for a description of a single service, we expect
    # there to only be a single service.
    services = resp['services']
    assert len(services) == 1, resp
    service = services[0]

    # Within a 'service' description, the following structure is what we're
    # interested in:
    #
    #     "deployments": [
    #       {
    #         "createdAt": <date>,
    #         "taskDefinition": <task definition ARN>,
    #         "updatedAt": <date>
    #         ...
    #       },
    #       ... other running tasks
    #     ],
    #
    # Each "deployment" corresponds to a running task, so we pick the
    # container with the most recent creation date.
    deployments = service['deployments']
    assert len(deployments) > 0, resp
    newest_deployment = max(deployments, key=operator.itemgetter('createdAt'))

    return newest_deployment['taskDefinition']


def clone_task_definition(client, task_definition):
    """
    Given a task definition ARN, clone the associated task.

    Returns the new task definition ARN.
    """
    resp = client.describe_task_definition(taskDefinition=task_definition)
    taskDefinition = resp['taskDefinition']

    # The task definition contains two key fields: "family" and
    # "containerDefinitions" which full describe the task.
    new_task = client.register_task_definition(
        family=taskDefinition['family'],
        taskRoleArn=taskDefinition['taskRoleArn'],
        containerDefinitions=taskDefinition['containerDefinitions'],
        volumes=taskDefinition['volumes']
    )

    return new_task['taskDefinition']['taskDefinitionArn']


def _name_from_arn(arn):
    return arn.split("/")[1]


def _check_for_throttle_exception(f, *args, **kwargs):
    try:
        return f(*args, **kwargs)
    except ClientError as ex:
        if ex.response['Error']['Code'] == 'ThrottlingException':
            print(f'ThrottlingException: {ex}')
            raise EcsThrottleException(ex)
        else:
            raise


def get_service_arns(ecs_client, cluster_arn):
    """
    Given a cluster ARN, extracts the associated service ARNs.

    Returns a list of service ARNS.
    """
    return _check_for_throttle_exception(
        ecs_client.list_services,
        cluster=_name_from_arn(cluster_arn)
    )['serviceArns']


def get_cluster_arns(ecs_client):
    """
    Extract the list of cluster ARNs in this account.

    Returns a list of cluster ARNs.
    """
    return _check_for_throttle_exception(
        ecs_client.list_clusters
    )['clusterArns']


def describe_cluster(ecs_client, cluster_arn):
    """
    Given a cluster ARN attempts to find a matching cluster description.

    Returns a cluster description.
    """
    return _check_for_throttle_exception(
        ecs_client.describe_clusters,
        clusters=[cluster_arn]
    )['clusters'][0]


def describe_service(ecs_client, cluster_arn, service_arn):
    """
    Given a cluster ARN and service ARN, attempts to find a matching
    service description.

    Returns a service description.
    """
    return _check_for_throttle_exception(
        ecs_client.describe_services,
        cluster=_name_from_arn(cluster_arn),
        services=[_name_from_arn(service_arn)]
    )['services'][0]


def run_task(
        ecs_client,
        cluster_name,
        task_definition,
        started_by,
        container_name="app",
        command=[]):
    """
    Run a given command against a named container in a task definition
    on a particular cluster.

    Returns the response from calling run_task
    """
    return ecs_client.run_task(
        cluster=cluster_name,
        taskDefinition=task_definition,
        overrides={
            'containerOverrides': [
                {
                    'name': container_name,
                    'command': command
                },
            ]
        },
        count=1,
        startedBy=started_by,
    )



# -*- encoding: utf-8 -*-


def _is_capacity_different(x, desired_capacity):
    read_capacity_units = x['ProvisionedThroughput']['ReadCapacityUnits']
    write_capacity_units = x['ProvisionedThroughput']['WriteCapacityUnits']
    return (
        read_capacity_units != desired_capacity
    ) or (
        write_capacity_units != desired_capacity
    )


def change_dynamo_capacity(client, table_name, desired_capacity):
    """
    Given the name of a DynamoDB table and a desired capacity, update the
    read/write capacity of the table and every secondary index.
    """

    response = client.describe_table(TableName=table_name)

    filtered_gsis = filter(
        lambda x: _is_capacity_different(x, desired_capacity),
        response['Table']['GlobalSecondaryIndexes'])

    gsi_updates = list(map(
        lambda x: {
            'Update': {
                'IndexName': x['IndexName'],
                'ProvisionedThroughput': {
                    'ReadCapacityUnits': desired_capacity,
                    'WriteCapacityUnits': desired_capacity
                }
            }
        },
        filtered_gsis
    ))

    table_update = _is_capacity_different(response['Table'], desired_capacity)
    print(f'table_update: {table_update}')

    if gsi_updates and table_update:
        resp = client.update_table(
            TableName=table_name,
            ProvisionedThroughput={
                'ReadCapacityUnits': desired_capacity,
                'WriteCapacityUnits': desired_capacity
            },
            GlobalSecondaryIndexUpdates=gsi_updates
        )
    elif gsi_updates:
        resp = client.update_table(
            TableName=table_name,
            GlobalSecondaryIndexUpdates=gsi_updates
        )
    elif table_update:
        resp = client.update_table(
            TableName=table_name,
            ProvisionedThroughput={
                'ReadCapacityUnits': desired_capacity,
                'WriteCapacityUnits': desired_capacity
            }
        )
    else:
        return

    print(f'DynamoDB response = {resp!r}')
    assert resp['ResponseMetadata']['HTTPStatusCode'] == 200



# -*- encoding: utf-8 -*-

import boto3
import daiquiri


logger = daiquiri.getLogger(__name__)


def get_messages(queue_url, delete=False, batch_size=10):
    """
    Gets messages from an SQS queue.  If ``delete`` is True, the
    messages are also deleted after they've been read.
    """
    client = boto3.client('sqs')
    while True:
        # We batch message responses to reduce load on the SQS API.
        # Note: 10 is currently the most messages you can read at once.
        resp = client.receive_message(
            QueueUrl=queue_url,
            AttributeNames=['All'],
            MaxNumberOfMessages=batch_size
        )

        # If there's nothing available, the queue is empty.  Abort!
        try:
            logger.info(
                'Received %d new messages from %s',
                len(resp['Messages']), queue_url)
        except KeyError:
            logger.info('No messages received from %s; aborting', queue_url)
            break

        # If we're deleting the messages ourselves, we don't need to send
        # the ReceiptHandle to the caller (it's only used for deleting).
        # If not, we send the entire response.
        if delete:
            for m in resp['Messages']:
                yield {k: v for k, v in m.items() if k != 'ReceiptHandle'}
        else:
            yield from resp['Messages']

        # Now delete the messages from the queue, so they won't be read
        # on the next GET call.
        if delete:
            logger.info(
                'Deleting %d messages from %s',
                len(resp['Messages']), queue_url)
            client.delete_message_batch(
                QueueUrl=queue_url,
                Entries=[
                    {'Id': m['MessageId'], 'ReceiptHandle': m['ReceiptHandle']}
                    for m in resp['Messages']
                ]
            )



# -*- encoding: utf-8

import functools
import sys


def log_on_error(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            print(f'args   = {args!r}', file=sys.stderr)
            print(f'kwargs = {kwargs!r}', file=sys.stderr)
            raise

    return wrapper



# -*- encoding: utf-8 -*-

from wellcome_aws_utils import (
    deployment_utils,
    dynamo_utils,
    ecs_utils,
    s3_utils,
    sns_utils,
    sqs_utils,
)
from wellcome_aws_utils.version import __version_info__, __version__


__all__ = [
    '__version_info__',
    '__version__',
    'deployment_utils',
    'dynamo_utils',
    'ecs_utils',
    's3_utils',
    'sns_utils',
    'sqs_utils',
]



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
Shared library to help surface ECS deployment information.
"""

import collections
import datetime

from wellcome_aws_utils.ecs_utils import (
    get_cluster_arns, get_service_arns, describe_service
)

Deployment = collections.namedtuple(
    'Deployment',
    'deployment_key deployment_status color created_at task_definition'
)
DeploymentKey = collections.namedtuple('DeploymentKey', 'id service_arn')


DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


def _create_deployment_tuple_from_item(item):
    item_date = datetime.datetime.strptime(item['created_at'], DATE_FORMAT)

    return Deployment(
        DeploymentKey(item['deployment_id'], item['service_arn']),
        item['deployment_status'],
        item['color'],
        item_date,
        item['task_definition']
    )


def _create_deployment_tuple_from_ecs(service, deployment):
    """Takes AWS ECS API Service & Deployment, return Deployment namedtuple."""
    deployment_status = deployment['status']
    ongoing_deployment = len(service['deployments']) > 1

    if ongoing_deployment and (deployment_status == "PRIMARY"):
        color = "green"
    else:
        color = "blue"

    return Deployment(
        DeploymentKey(deployment['id'], service['serviceArn']),
        deployment_status,
        color,
        deployment['createdAt'],
        deployment['taskDefinition']
    )


def _get_service_deployments(ecs_client, cluster_arn, service_arn):
    service = describe_service(ecs_client, cluster_arn, service_arn)

    return [_create_deployment_tuple_from_ecs(service, deployment)
            for deployment in service['deployments']]


def _get_date_string(date):
    return date.strftime(DATE_FORMAT)


def delete_deployment_in_dynamo(table, deployment):
    return table.delete_item(
        Key={
            'deployment_id': deployment.deployment_key.id,
            'service_arn': deployment.deployment_key.service_arn
        }
    )


def put_deployment_in_dynamo(table, deployment):
    return table.put_item(
        Item={
            'deployment_id': deployment.deployment_key.id,
            'service_arn': deployment.deployment_key.service_arn,
            'deployment_status': deployment.deployment_status,
            'color': deployment.color,
            'created_at': _get_date_string(deployment.created_at),
            'task_definition': deployment.task_definition
        }
    )


def update_deployment_in_dynamo(table, deployment):
    return table.update_item(
        Key={
            'deployment_id': deployment.deployment_key.id,
            'service_arn': deployment.deployment_key.service_arn
        },
        UpdateExpression="""
            SET deployment_status = :deployment_status,
                color = :color,
                created_at = :created_at,
                task_definition = :task_definition
        """,
        ExpressionAttributeValues={
            ':deployment_status': deployment.deployment_status,
            ':color': deployment.color,
            ':created_at': _get_date_string(deployment.created_at),
            ':task_definition': deployment.task_definition
        }
    )


def get_deployments_from_dynamo(table):
    response = table.scan()

    return [_create_deployment_tuple_from_item(d) for d in response['Items']]


def get_deployments_from_ecs(ecs_client):
    deployments = []

    for cluster_arn in get_cluster_arns(ecs_client):
        for service_arn in get_service_arns(ecs_client, cluster_arn):
            service_deployments = _get_service_deployments(
                ecs_client, cluster_arn, service_arn)
            deployments += service_deployments

    return deployments



# -*- encoding: utf-8 -*-

import os

from setuptools import find_packages, setup


def local_file(name):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), name))


SOURCE = local_file('src')
README = local_file('README.rst')


# Assignment to placate pyflakes. The actual version is from the exec that
# follows.
__version__ = None

with open(local_file('src/wellcome_aws_utils/version.py')) as o:
    exec(o.read())

assert __version__ is not None


setup(
    name='wellcome_aws_utils',
    packages=find_packages(SOURCE),
    package_dir={'': SOURCE},
    version=__version__,
    install_requires=[
        'boto3',
        'daiquiri',
        'python-dateutil',
        'elasticsearch',
        'attrs'
    ],
    python_requires='>=3.6',
    description='A collection of AWS utilities',
    long_description=open(README).read(),
    author='Wellcome Trust (Digital Platform Team)',
    author_email='wellcomedigitalplatform@wellcome.ac.uk',
    url='https://github.com/wellcometrust/aws_utils',
    keywords=['aws'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
)
